-- ============================================================= --
-- KEYBOARD STEERING MOD
-- ============================================================= --
KeyboardSteering = {};

addModEventListener(KeyboardSteering);

function KeyboardSteering.prerequisitesPresent(specializations)
	return  SpecializationUtil.hasSpecialization(Drivable, specializations) and
			SpecializationUtil.hasSpecialization(Motorized, specializations) and
			SpecializationUtil.hasSpecialization(Enterable, specializations)
end

function KeyboardSteering.registerEventListeners(vehicleType)
    table.insert(vehicleType.eventListeners.onEnterVehicle, KeyboardSteering)
    table.insert(vehicleType.eventListeners.onLeaveVehicle, KeyboardSteering)
    table.insert(vehicleType.eventListeners.onRegisterActionEvents, KeyboardSteering)
end

function KeyboardSteering:onRegisterActionEvents(isSelected, isOnActiveVehicle)
    local actionEventId
    if isOnActiveVehicle then
        _, actionEventId = InputBinding.registerActionEvent(g_inputBinding, 'KBSTEERING_ENABLE_DISABLE', self, KeyboardSteering.toggleEnable, false, true, false, true)
		g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_LOW)
		g_inputBinding:setActionEventTextVisibility(actionEventId, false)
    end
end
function KeyboardSteering:onEnterVehicle(isControlling, playerStyle, farmId)
	KeyboardSteering.displayToggleState = false
end
function KeyboardSteering:onLeaveVehicle(isControlling, playerStyle, farmId)
	KeyboardSteering.displayToggleState = false
end

function KeyboardSteering:toggleEnable(actionName, inputValue)
	if KeyboardSteering.isControledVehicle then
		if actionName=='KBSTEERING_ENABLE_DISABLE' and inputValue==1 then
			KeyboardSteering.displayToggleState = true
			if not KeyboardSteering.enable then
				--print("ENABLE")
				KeyboardSteering.enable = true
			else
				--print("DISABLE")
				KeyboardSteering.enable = false
			end
		end
	end
end

function KeyboardSteering:onUpdate(superFunc, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if not KeyboardSteering.enable then
		return superFunc(self, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	end
	
	--print("Modified Steering Function")
    local spec = self.spec_drivable
	
    -- update inputs on client side for basic controls
    if self.isClient then
        if self.getIsEntered ~= nil and self:getIsEntered() then
            if self:getIsActiveForInput(true, true) then
                if self:getIsVehicleControlledByPlayer() then
					KeyboardSteering.isControledVehicle = true
                    spec.doHandbrake = false
                    -- gas and brake pedal
                    local axisForward = MathUtil.clamp((spec.lastInputValues.axisAccelerate - spec.lastInputValues.axisBrake), -1, 1)
                    spec.axisForward = axisForward
                    -- steering
                    local speedFactor = 1.0
                    local sensitivitySetting = g_gameSettings:getValue(GameSettings.SETTING.STEERING_SENSITIVITY)
					
                    if spec.lastInputValues.axisSteerIsAnalog then
                        local isArticulatedSteering = self.spec_articulatedAxis ~= nil and self.spec_articulatedAxis.componentJoint ~= nil
                        if isArticulatedSteering then
                            speedFactor = 1.5
                        else
                            speedFactor = 2.5
                        end
                        -- only use steering speed for gamepads
                        if spec.lastInputValues.axisSteerDeviceCategory == InputDevice.CATEGORY.GAMEPAD then
                            speedFactor = speedFactor * sensitivitySetting
                        end
                    else
                        if spec.lastInputValues.axisSteer == 0 then
                            local rotateBackSpeedSetting = g_gameSettings:getValue(GameSettings.SETTING.STEERING_BACK_SPEED) / 10
                            if self.speedDependentRotateBack then
                                local speed = self:getLastSpeed()
                                local setting = rotateBackSpeedSetting / 0.5
								local maxSpeed = 50 --self:getMotor():getMaximumForwardSpeed()*3.6
								local roadSpeedFactor = math.min(speed/maxSpeed, 1.0)
								local steeringAngleFactor = math.abs(spec.axisSide)
								speedFactor = speedFactor * math.min(setting * roadSpeedFactor * steeringAngleFactor, 2)
                            end
                            speedFactor = speedFactor * (self.autoRotateBackSpeed or 1) / 1.5
                        else
							local keyPressTimeFactor = KeyboardSteering.timeKeyHeld/KeyboardSteering.maxKeyHeldTime
                            speedFactor = math.min(1.0/(self.lastSpeed*spec.speedRotScale+spec.speedRotScaleOffset), 1);
                            speedFactor = speedFactor * sensitivitySetting * keyPressTimeFactor
                        end
                    end
                    local steeringDuration = (self.wheelSteeringDuration or 1) * 1000
                    local rotDelta = (dt / steeringDuration) * speedFactor
					
					-- REMOVE ALWAYS TURNING RIGHT BEHAVIOUR WHEN HOLDING DOWN BOTH KEYS
					if KeyboardSteering.goStraight then
						--print("go straight")
						if math.abs(spec.axisSide)<0.01 then
							spec.axisSide = 0
							rotDelta = 0
						else
							if spec.axisSide>0.0 then
								if spec.axisSide==1.0 then
									spec.axisSide=spec.axisSide-rotDelta
								end
								rotDelta = -rotDelta
							end
						end
					end
					
					if spec.lastInputValues.axisSteer > spec.axisSide then
                        spec.axisSide = math.min(spec.lastInputValues.axisSteer, spec.axisSide + rotDelta)
                    elseif spec.lastInputValues.axisSteer < spec.axisSide then
                        spec.axisSide = math.max(spec.lastInputValues.axisSteer, spec.axisSide - rotDelta)
                    end
					--Limit steering angle depending on road speed
					local speed = self:getLastSpeed()
					local rotateBackSetting = g_gameSettings:getValue(GameSettings.SETTING.STEERING_BACK_SPEED) / 10
					rotateBackSetting = MathUtil.clamp(rotateBackSetting, 0, 1)
					local minSpeed = 50 - (30*rotateBackSetting)
					local maxSpeed = 500 - (360*rotateBackSetting)
					if speed>minSpeed then
						local maxAngle = MathUtil.clamp(1.0 - (speed-minSpeed)/(maxSpeed-minSpeed), 0.2, 1)
						spec.axisSide = MathUtil.clamp(spec.axisSide, -maxAngle, maxAngle)
					end
                else
					KeyboardSteering.isControledVehicle = false
                    spec.axisForward = 0
                    if self.rotatedTime < 0 then
                        spec.axisSide = self.rotatedTime / -self.maxRotTime / self:getSteeringDirection()
                    else
                        spec.axisSide = self.rotatedTime / self.minRotTime / self:getSteeringDirection()
                    end
                end
            else
                spec.doHandbrake = true
                spec.axisForward = 0
            end
            -- prepare for next frame
            spec.lastInputValues.axisAccelerate = 0
            spec.lastInputValues.axisBrake = 0
            spec.lastInputValues.axisSteer = 0
            -- prepare network update
            if spec.axisForward ~= spec.axisForwardSend or spec.axisSide ~= spec.axisSideSend or spec.doHandbrake ~= spec.doHandbrakeSend then
                spec.axisForwardSend = spec.axisForward
                spec.axisSideSend = spec.axisSide
                spec.doHandbrakeSend = spec.doHandbrake
                self:raiseDirtyFlags(spec.dirtyFlag)
            end
        end
    end
    -- update inputs on client side for cruise control
    if self.isClient then
        if self.getIsEntered ~= nil and self:getIsEntered() then
            -- cruise control state
            local inputValue = spec.lastInputValues.cruiseControlState
            spec.lastInputValues.cruiseControlState = 0
            if inputValue == 1 then
                if spec.cruiseControl.topSpeedTime == Drivable.CRUISECONTROL_FULL_TOGGLE_TIME then
                    if spec.cruiseControl.state == Drivable.CRUISECONTROL_STATE_OFF then
                        self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_ACTIVE)
                    else
                        self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_OFF)
                    end
                end
                if spec.cruiseControl.topSpeedTime > 0 then
                    spec.cruiseControl.topSpeedTime = spec.cruiseControl.topSpeedTime - dt
                    if spec.cruiseControl.topSpeedTime < 0 then
                        self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_FULL)
                    end
                end
            else
                spec.cruiseControl.topSpeedTime = Drivable.CRUISECONTROL_FULL_TOGGLE_TIME
            end
            -- cruise control value
            local lastCruiseControlValue = spec.lastInputValues.cruiseControlValue
            spec.lastInputValues.cruiseControlValue = 0
            if lastCruiseControlValue ~= 0 then
                spec.cruiseControl.changeCurrentDelay = spec.cruiseControl.changeCurrentDelay - (dt * spec.cruiseControl.changeMultiplier)
                spec.cruiseControl.changeMultiplier = math.min(spec.cruiseControl.changeMultiplier + (dt * 0.003), 10)
                if spec.cruiseControl.changeCurrentDelay < 0 then
                    spec.cruiseControl.changeCurrentDelay = spec.cruiseControl.changeDelay
                    local dir = MathUtil.sign(lastCruiseControlValue)
                    local speed = spec.cruiseControl.speed + dir
                    self:setCruiseControlMaxSpeed(speed)
                    if spec.cruiseControl.speed ~= spec.cruiseControl.speedSent then
                        if g_server ~= nil then
                            g_server:broadcastEvent(SetCruiseControlSpeedEvent:new(self, spec.cruiseControl.speed), nil, nil, self)
                        else
                            g_client:getServerConnection():sendEvent(SetCruiseControlSpeedEvent:new(self, spec.cruiseControl.speed))
                        end
                        spec.cruiseControl.speedSent = spec.cruiseControl.speed
                    end
                end
            else
                spec.cruiseControl.changeCurrentDelay = 0
                spec.cruiseControl.changeMultiplier = 1
            end
        end
    end
    local isControlled = self.getIsControlled ~= nil and self:getIsControlled()
    -- update vehicle physics on server side
    if self:getIsVehicleControlledByPlayer() then
        if self.isServer then
            if isControlled then
                -- lock max speed to working tool
                local speed,_ = self:getSpeedLimit(true)
                if spec.cruiseControl.state == Drivable.CRUISECONTROL_STATE_ACTIVE then
                    speed = math.min(speed, spec.cruiseControl.speed)
                end
                self:getMotor():setSpeedLimit(speed)
                self:updateVehiclePhysics(spec.axisForward, spec.axisSide, spec.doHandbrake, dt)
            end
        end
    end
    -- just a visual update of the steering wheel
    if self.isClient and isControlled then
        self:updateSteeringWheel(spec.steeringWheel, dt, 1)
    end
end

function KeyboardSteering:loadMap(name)
	--print("Load Mod: 'Keyboard Steering'")
	Drivable.onUpdate = Utils.overwrittenFunction(Drivable.onUpdate, KeyboardSteering.onUpdate)

	KeyboardSteering.enable					= true		-- Enabled/Disabled
	KeyboardSteering.turnLeft				= false		-- input left turn(via key press)
	KeyboardSteering.turnRight				= false		-- input right turn (via key press)
	KeyboardSteering.timeKeyHeld			= 0			-- key down time for any steering input
	KeyboardSteering.maxKeyHeldTime			= 500		-- maximum key down time (ms)
	KeyboardSteering.goStraight				= false		-- flag to go straight after pressing left+right
	KeyboardSteering.goStraightReleaseTime 	= 0			-- both keys down released time
	KeyboardSteering.goStraightReleaseDelay = 100		-- both keys down release delay
	KeyboardSteering.isControledVehicle		= false		-- flag for steering key inputs to be captured
	KeyboardSteering.displayToggleState 	= false		-- show toggled state in F1 menu
	KeyboardSteering.initialised			= false		-- inisitalised flag
end

function KeyboardSteering:deleteMap()
end

function KeyboardSteering:mouseEvent(posX, posY, isDown, isUp, button)
end

function KeyboardSteering:keyEvent(unicode, sym, modifier, isDown)
	if KeyboardSteering.isControledVehicle then
		for _, data in pairs(KeyboardSteering.leftKey) do
			if sym==data.id then
				if isDown then
					KeyboardSteering.turnLeft = true
				else
					KeyboardSteering.turnLeft = false
				end
				return
			end
		end
		for _, data in pairs(KeyboardSteering.rightKey) do
			if sym==data.id then
				if isDown then
					KeyboardSteering.turnRight = true
				else
					KeyboardSteering.turnRight = false
				end
				return
			end
		end
	end
end

function KeyboardSteering:draw()
	if KeyboardSteering.displayToggleState then
		if KeyboardSteering.enable then
			g_currentMission:addExtraPrintText("Keyboard Steering: Enabled")
		else
			g_currentMission:addExtraPrintText("Keyboard Steering: Disabled")
		end
	end
end

function KeyboardSteering:update(dt)
	if not KeyboardSteering.initialised then
		if g_settingsScreen.settingsModel.steeringBackSpeedValues[12] == nil then
			--EXTEND RANGE OF STEERING BACK SPEED SETTINGS
			local newValues = {11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0}
			local newStrings = {'110%', '120%', '130%', '140%', '150%', '160%', '170%', '180%', '190%', '200%'}
			for i = 1, 10 do
				table.insert(g_settingsScreen.settingsModel.steeringBackSpeedValues, newValues[i])
				table.insert(g_settingsScreen.settingsModel.steeringBackSpeedStrings, newStrings[i])
			end
		end

		--DEFAULT KEYS FOR STEERING
		KeyboardSteering.leftKey = {}
		KeyboardSteering.rightKey = {}
		--DETECT BOUND KEYS FOR STEERING	
		local xmlFile = loadXMLFile('TempXML', g_gui.inputManager.settingsPath)	
		local actionBindingCounter = 0
		if xmlFile ~= 0 then
			while true do
				local key = string.format('inputBinding.actionBinding(%d)', actionBindingCounter)
				local actionString = getXMLString(xmlFile, key .. '#action')
				if actionString == nil then
					break
				end
				if actionString == 'AXIS_MOVE_SIDE_VEHICLE' then
					--print( actionString )
					local i = 0
					while true do
						local bindingKey = key .. string.format('.binding(%d)',i)
						local bindingName = getXMLString(xmlFile, bindingKey .. '#input')
						local bindingIndex = getXMLString(xmlFile, bindingKey .. '#index')
						local bindingAxisComponent = getXMLString(xmlFile, bindingKey .. '#axisComponent')
						if bindingName == nil then
							break
						end
						
						--print( bindingName )
						if bindingAxisComponent == '-' then
							table.insert(KeyboardSteering.leftKey, {name=bindingName, id=Input[bindingName], index=bindingIndex })
						else
							table.insert(KeyboardSteering.rightKey, {name=bindingName, id=Input[bindingName], index=bindingIndex })
						end
						i = i + 1
					end
				end
				actionBindingCounter = actionBindingCounter + 1
			end
		end
		delete(xmlFile)
				
		if next(KeyboardSteering.leftKey)==nil or next(KeyboardSteering.rightKey)==nil then
			print("Keyboard Steering: no keys bound to vehicle turn left/right")
			KeyboardSteering.enable = false
			--table.insert(KeyboardSteering.leftKey, {name='KEY_a', id=Input['KEY_a'], index=1 })
			--table.insert(KeyboardSteering.rightKey, {name='KEY_d', id=Input['KEY_d'], index=1 })
		end
			
		KeyboardSteering.initialised = true
	end
	
	
	-- UPDATE KEY PRESS TIMES AND FLAGS
	if KeyboardSteering.turnLeft and KeyboardSteering.turnRight then
		-- GOING STRAIGHT
		KeyboardSteering.goStraight = true
		KeyboardSteering.timeKeyHeld = KeyboardSteering.timeKeyHeld + dt
		KeyboardSteering.goStraightReleaseTime = 0
	else
		if KeyboardSteering.turnLeft or KeyboardSteering.turnRight then
			-- TURNING LEFT OR RIGHT
			KeyboardSteering.timeKeyHeld = KeyboardSteering.timeKeyHeld + dt

			-- CREATE DEADTIME AFTER RELEASING BOTH KEYS
			if KeyboardSteering.goStraight then
				KeyboardSteering.goStraightReleaseTime = KeyboardSteering.goStraightReleaseTime + dt
				if KeyboardSteering.goStraightReleaseTime > KeyboardSteering.goStraightReleaseDelay then
					KeyboardSteering.goStraight = false
				end
			end
		else
			-- no keyboard input
			KeyboardSteering.timeKeyHeld = KeyboardSteering.timeKeyHeld - dt
			KeyboardSteering.goStraight = false
		end
		
	end
	KeyboardSteering.timeKeyHeld = MathUtil.clamp(KeyboardSteering.timeKeyHeld, 0, KeyboardSteering.maxKeyHeldTime)
end